/*
// UNIT TESTS OF THE BOT
                               
                                                   brake  M1   brake  M2   brake  M3    LACT
INPUT: FULL BRAKE Command String                = ~+0001,+0000,+0001,+0000,+0001,+0000,+0000
       Release Brakes Command String            = ~+0000,+0000,+0000,+0000,+0000,+0000,+0000
       Full Motor Forward String                = ~+0000,+0100,+0000,+0100,+0000,+0000,+0000
       Full Motor Backward String               = ~+0000,-0100,+0000,-0100,+0000,+0000,+0000

       Turn Right                               = ~+0000,+0100,+0000,-0100,+0000,+0000,+0000
       Turn Left                                = ~+0000,-0100,+0000,+0100,+0000,+0000,+0000
       
       Left Motor Command String                = ~+0000,+0100,+0000,+0000,+0000,+0000,+0000
       Right Motor Command String               = ~+0000,+0000,+0000,+0100,+0000,+0000,+0000
       Flipper forwards MotorCommand String     = ~+0000,+0000,+0000,+0000,+0000,+0100,+0000
       Flipper backwards MotorCommand String    = ~+0000,+0000,+0000,+0000,+0000,-0100,+0000
       LACT forwards MotorCommand String        = ~+0000,+0000,+0000,+0000,+0000,+0000,+3200
       LACT backwards MotorCommand String       = ~+0000,+0000,+0000,+0000,+0000,+0000,-3200


OUTPUT: ir sensor array in cm : (fl, fr, l, r, bl, br) from front left to back right
                      IMU Data:
                                IMU Pulse
                                ax = 2.81 ay = -66.22 az = -998.54 mg
                                gx = 0.02 gy = 0.05 gz = -0.08 deg/s
                                mx = -1205 my = -1865 mz = 1386 mG
                                q0 = -0.04 qx = -0.06 qy = -0.53 qz = -0.85
                                Yaw, Pitch, Roll: 168.17, -3.40, 64.22
                                Temperature is 27.6 degrees C
                                rate = 0.23 Hz

--------- NEW TARGET Motor Values -----------
target_Motor1_Command: 0
target_Motor2_Command: 0
target_FlipperMotor_Command: 0
target_LACT_Command: 0
---------------------------------------
blip - 60 milliseconds elapsed --- ( 97, 68, 129, 204, 190, 83)


TODO: animate motor values to the specified number on arduino with saftey animations
        • currently only when the serial pulse comes in will the motors be set
            I need to keep the motor animations going until they are finished animating in a separate loop from serial parsing
                    1. after serial pulse parse, we should have some kind of updated target values to animate to
                    2.  problem: the SAFTEY STOP SHUTDOWN code needs to be modified and tested very thuroughly since animating target values will 
                                  be affecting the target value to animate... 
                        solution: the safty shutdown needs to animate the target motor values and the animation should 
                                  finish animations to the next target value which will shutoff at 0... but not abruptly right? just eased stop 
                                  minimizing wear on parts
                    2. in the regular loop by the IMU processing code, find a place to check if targetValuesNeedAnimating
                        if (Yes) then we animate the motor values until we are at the target value
                        if (No) continue...
                  we could normalize the speed from 0-590 with 0 at full reverse and 590 at full forward, animating in between

                  *** Rewire 
                  *   A8 - D18 -> interrupt5 ----- M1 Speed Sensor ( Left Motor )
                  *   A9 - D19 -> interrupt4 ----- M2 Speed Sensor ( Right Motor )
                  *   A10 - D20 -> interrupt3 ---- Flipper Speed Sensor
           ****************************************************************************************************
           *    This only accounts for the animation of the forward or backward directions... 
           *    in reality this value will need animating in the reverse
           *    direction to properly brake the robot and then reverse..... 
           *    full forward to full backward with a time interval for transition  
           *    Make sure to check for direction as well: ex. 255Forw->255Back needs full animation
           ****************************************************************************************************
            
      • Unify the output to the computer so that IMU and IRSensor array is returned in a modular logical way
      • I need to get the speed control output working... https://brainy-bits.com/blogs/tutorials/speed-sensor-with-arduino
            set the digital speed control input pins and find a library to get motor skips somehow
          "We will use interrupts and the TimerOne library in our code. "
          there you have it... in needs to be on pin2 for the UNO because it has an interrupt capability
          I need to check what I did with the digital pin hookup and see if its capable of the interrupt using TimerOne library
      • I need the voltage testing opamp circuit to be implemented 
          http://www.skillbank.co.uk/arduino/measure2.htm
          http://www.skillbank.co.uk/arduino/calibrate.htm
          
*/
//----- IMU Hookups -----
#include "quaternionFilters.h"
#include "MPU9250.h"

MPU9250 myIMU;
#define IMU_PROCESSING_SPEED1 10000
#define IMU_PROCESSING_SPEED2 50        // 100 = .12 hz       50 should be = 0.06 hz (6-fps)
int IMU_Processing_Index1 = 0; //Clock is so fast we need 2 variables to choke the processing... im talkin MHz fast 
int IMU_Processing_Index2 = 0; //second choke variables that starts to turn nano seconds into mili seconds

//----- LACT Hookups -----
#include <SoftwareSerial.h>
#define rxPin 22  // pin 3 connects to smcSerial TX  (not used in this example)
#define txPin 23  // pin 4 connects to smcSerial RX
SoftwareSerial smcSerial = SoftwareSerial(rxPin, txPin);


//----- Motor Hookups -----
#define TRED_R_MOTOR_SPEED_PIN       2
#define TRED_R_MOTOR_BRAKE_PIN       3
#define TRED_R_MOTOR_DIRECTION_PIN   25
//TODO: SpeedSensor - #define TRED_R_MOTOR_SPEEDSENSOR_PIN 18 //interrupt5

#define TRED_L_MOTOR_SPEED_PIN       4
#define TRED_L_MOTOR_BRAKE_PIN       5
#define TRED_L_MOTOR_DIRECTION_PIN   27
//TODO: SpeedSensor - #define TRED_L_MOTOR_SPEEDSENSOR_PIN 19 //interrupt4

#define FLIPPER_MOTOR_SPEED_PIN       6
#define FLIPPER_MOTOR_BRAKE_PIN       7
#define FLIPPER_MOTOR_DIRECTION_PIN   29
//TODO: SpeedSensor - #define FLIPPER_MOTOR_SPEEDSENSOR_PIN 20 //interrupt3

#define RAMP_SPEED_SLOW 20
#define RAMP_SPEED_MED 10
#define RAMP_SPEED_FAST 1

//----- DistanceSaftey Defines -----
#define FORWARD_SAFE_DISTANCE     25
#define BACKWARD_SAFE_DISTANCE    25

//----- KeepAlive Variables -----
#define KEEP_ALIVE_TIMEOUT_1      10000 
#define KEEP_ALIVE_TIMEOUT_2      4 ////roughly every 2 seconds if we use 20 but lets try 2
unsigned long start, finished, elapsed;
int keepAlive_1 = KEEP_ALIVE_TIMEOUT_1;
int keepAlive_2 = KEEP_ALIVE_TIMEOUT_2;

int SAFTEY_SHUTDOWN_FORWARD = 0;
int SAFTEY_SHUTDOWN_BACKWARD = 0;

//----- SerialComms Variables -----
char targetCommandString[] = {'0','0','0','0','0'};

int target_Motor1_Command = 0;
int target_Motor1_brake_command = 0;
int target_Motor2_Command = 0;
int target_Motor2_brake_command = 0;
int target_FlipperMotor_Command = 0;
int target_FlipperMotor_brake_command = 0;
int target_LACT_Command = 0;

//----- IR BLOCK -----
#include <SharpIR.h>

int IRMAXSHUTDOWNCOUNT           = 10;
int currentIRSHUTDOWNCOUNT_FRONT = 0;
int currentIRSHUTDOWNCOUNT_BACK  = 0;

#define IR_FRONT_LEFT             A5
#define IR_FRONT_RIGHT            A4
#define IR_LEFT                   A3
#define IR_RIGHT                  A2  
#define IR_BACK_LEFT              A1
#define IR_BACK_RIGHT             A0

#define LONG_RANGE_IR_SENSOR      20150
#define MED_RANGE_IR_SENSOR       1080

SharpIR IR_FL(IR_FRONT_LEFT, LONG_RANGE_IR_SENSOR);
SharpIR IR_FR(IR_FRONT_RIGHT, LONG_RANGE_IR_SENSOR);
SharpIR IR_L(IR_LEFT, MED_RANGE_IR_SENSOR);
SharpIR IR_R(IR_RIGHT, MED_RANGE_IR_SENSOR);
SharpIR IR_BL(IR_BACK_LEFT, LONG_RANGE_IR_SENSOR);
SharpIR IR_BR(IR_BACK_RIGHT, LONG_RANGE_IR_SENSOR);
//--------------------


void setup() {
  // put your setup code here, to run once:
  Wire.begin();

  Serial.begin( 250000 );
  
  //----- IMU SETUP -----
  initialize_IMU();

  //----- LACT SETUP -----
  smcSerial.begin(19200);
  delay(5);
  smcSerial.write(0xAA);  // send baud-indicator byte
  exitSafeStart();  // clear the safe-start violation and let the motor run

 
  Serial.println("iBot v1.0 by OrbitusRobotics");
  Serial.println("RHAPI v0.1");
  Serial.println("BEGIN BASE STARTUP SEQUENCE");
  //---- MotorPin SETUP ----
  pinMode( TRED_R_MOTOR_SPEED_PIN, OUTPUT);   
  pinMode( TRED_R_MOTOR_BRAKE_PIN, OUTPUT);
  pinMode( TRED_R_MOTOR_DIRECTION_PIN, OUTPUT);   
  //TODO: SpeedSensor - pinMode( TRED_R_MOTOR_SPEEDSENSOR_PIN, INPUT);   
  
  pinMode( TRED_L_MOTOR_SPEED_PIN, OUTPUT);   
  pinMode( TRED_L_MOTOR_BRAKE_PIN, OUTPUT);
  pinMode( TRED_L_MOTOR_DIRECTION_PIN, OUTPUT);   
  //TODO: SpeedSensor - pinMode( TRED_L_MOTOR_SPEEDSENSOR_PIN, INPUT);   

  pinMode( FLIPPER_MOTOR_SPEED_PIN, OUTPUT);   
  pinMode( FLIPPER_MOTOR_BRAKE_PIN, OUTPUT);
  pinMode( FLIPPER_MOTOR_DIRECTION_PIN, OUTPUT);   
  //TODO: SpeedSensor - pinMode( FLIPPER_MOTOR_SPEEDSENSOR_PIN, INPUT);   

  //--------
  //ZERO THE SPEED
  analogWrite(TRED_L_MOTOR_SPEED_PIN, 255);
  analogWrite(TRED_R_MOTOR_SPEED_PIN, 255);
  //THEN RELEASE THE BRAKE
  releaseBrake();
  //--------

  
  
  //---- IR BLOCK SETUP ----
  pinMode( IR_FRONT_LEFT, INPUT);
  pinMode( IR_FRONT_RIGHT, INPUT);
  pinMode( IR_LEFT, INPUT);
  pinMode( IR_RIGHT, INPUT);
  pinMode( IR_BACK_LEFT, INPUT);
  pinMode( IR_BACK_RIGHT, INPUT);
  //------------------------

  Serial.println("STARTUP SEQUENCE COMPLETE");
  Serial.println("iBot v1.0 ONLINE");
 
  start=millis();

}

void loop() {
  
    keepAlive_SAFTEYSTOP();
    processIMU();
    
    if(Serial.available() > 0) {

      Serial.print("Parsing Serial and resetting keepAlive\n");
      resetKeepAlive();
         
      byte b =  Serial.read();
      if (b == 126)
      {
        char charIn = 0;

        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        target_Motor1_brake_command = atoi(targetCommandString);
        Serial.print("target_Motor1_brake_command ");
        Serial.println(target_Motor1_brake_command );

        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!

  //M1 - LEFT TRED        
        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        Serial.print("TargetCommandString M1:");
        Serial.println( targetCommandString );
        target_Motor1_Command = atoi(targetCommandString);

        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!
  //M2 - RIGHT TRED
        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        target_Motor2_brake_command = atoi(targetCommandString);

        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!
        
        Serial.print("target_Motor2_brake_command ");
        Serial.println(target_Motor2_brake_command );
        
        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        Serial.print("TargetCommandString M2:");
        Serial.println( targetCommandString );
        
        target_Motor2_Command = atoi(targetCommandString);

        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!
  //FLIPPER MOTOR
        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        target_FlipperMotor_brake_command = atoi(targetCommandString);
        
        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!
        
        Serial.print("target_FlipperMotor_brake_command ");
        Serial.println(target_FlipperMotor_brake_command );
        
        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();

        Serial.print("TargetCommandString FLipper:");
        Serial.println( targetCommandString );
        target_FlipperMotor_Command = atoi(targetCommandString);

        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!

  //LACT
        targetCommandString[0] = Serial.read();
        targetCommandString[1] = Serial.read();
        targetCommandString[2] = Serial.read();
        targetCommandString[3] = Serial.read();
        targetCommandString[4] = Serial.read();
  
        Serial.read(); //READ THE COMMA SEPARATING THINGS!!!
        //Serial.print("TargetCommandString LACT:");
        //Serial.print( targetCommandString );
        target_LACT_Command = atoi(targetCommandString);

        
  
        setMotorCommand(target_Motor1_brake_command, target_Motor1_Command, target_Motor2_brake_command, target_Motor2_Command, target_FlipperMotor_brake_command, target_FlipperMotor_Command, target_LACT_Command);
        
        //irReadOut();
        
        Serial.println("---------------------------------------");
        
      }
   }
}


int irReadOut_SAFTEYSTOP()
{
  /* Check for obstacles 
  Serial.print( "( " );
  Serial.print( IR_FL.distance() );
  Serial.print( ", " );
  Serial.print( IR_FR.distance() );
  Serial.print( ", " );
  Serial.print( IR_L.distance() );
  Serial.print( ", " );
  Serial.print( IR_R.distance() );
  Serial.print( ", " );
  Serial.print( IR_BL.distance() );
  Serial.print( ", " );
  Serial.print( IR_BR.distance() );
  Serial.print( ")\n" );
  */

      //if (currentIRSHUTDOWNCOUNT >= IRMAXSHUTDOWNCOUNT)
     // {
      //  irReadOut_SAFTEYSTOP(); //THIS IS THE CONTROL POINT FOR THE IR SHUTTING DOWN THE MOVEMENT
      //}
      
  //FORWARD SAFTEY CHECKS
  if ( IR_FL.distance() < FORWARD_SAFE_DISTANCE ||
       IR_FR.distance() < FORWARD_SAFE_DISTANCE)
  {
    Serial.println( "WARNING! FRONT" );
    currentIRSHUTDOWNCOUNT_FRONT++;
    if (currentIRSHUTDOWNCOUNT_FRONT >= IRMAXSHUTDOWNCOUNT)
    {
      Serial.println( "SAFTEY SHUTDOWN ERROR! OBSTACLE IS BLOCKING FRONT!" );
      //*****
      //SAFTEY_SHUTDOWN_FORWARD = 1;
      //SEND A WARNING TO THE CONTROLLER INSTEAD
      //*****
    }
  }
  else
  {
    SAFTEY_SHUTDOWN_FORWARD = 0;
    currentIRSHUTDOWNCOUNT_FRONT = 0;
  }
  
  //BACKWARD SAFTEY CHECKS
  if ( IR_BL.distance() < BACKWARD_SAFE_DISTANCE  ||
       IR_BR.distance() < BACKWARD_SAFE_DISTANCE )
  {
    Serial.println( "WARNING! BACK" );
    currentIRSHUTDOWNCOUNT_BACK++;
    if (currentIRSHUTDOWNCOUNT_BACK >= IRMAXSHUTDOWNCOUNT)
    {
      Serial.println( "SAFTEY SHUTDOWN ERROR! OBSTACLE IS BLOCKING BACK!" );
      //*****
      //SAFTEY_SHUTDOWN_BACKWARD = 1;
      //SEND A WARNING TO THE CONTROLLER INSTEAD
      //*****
    }
  }
  else
  {
    SAFTEY_SHUTDOWN_BACKWARD = 0;
    currentIRSHUTDOWNCOUNT_BACK = 0;
  }
  
  
  return 1;
}

//------ KEEP ALIVE FUNCTIONS ------

void keepAlive_SAFTEYSTOP()
{
  keepAlive_1--;
    if (keepAlive_1 <= 0)
    {
      //MAKE SURE TO SET MOTORS BACK TO ZERO AFTER THE TIMEOUT IN CASE OF FAILURE AT FULL SPEED
      keepAlive_1 = KEEP_ALIVE_TIMEOUT_1;
      keepAlive_2--;

      //Serial.print("blip - ");
      //irReadOut();
      //Serial.print(elapsed);
      //Serial.print(" milliseconds elapsed --- ");
      
      //Serial.println(keepAlive_2);
      
      //------------
      //TODO: fix this to be controlled by input settings and physically occlude the floor IR return noise with wooden blockers
      irReadOut_SAFTEYSTOP(); //THIS IS THE CONTROL POINT FOR THE IR SHUTTING DOWN THE MOVEMENT
      //------------
    }
    if (keepAlive_2 <= 0)
    {
      finished=millis();
      elapsed=finished-start;
      start=millis();
      Serial.print("KEEPALIVE HAS EXPIRED!!! SHUTTING DOWN CONNECTION NOW -- ");
      Serial.print(elapsed);
      Serial.println(" milliseconds");
      keepAlive_2 = KEEP_ALIVE_TIMEOUT_2;
      
      setMotorCommand(0,+0000, 0,+0000 ,0,+0000, +0000);
    }
}

void resetKeepAlive()
{
  keepAlive_1 = KEEP_ALIVE_TIMEOUT_1;
  keepAlive_2 = KEEP_ALIVE_TIMEOUT_2;
}

//------ MOTOR FUNCTIONS ------

void setMotorCommand(int target_Motor1_Brake_command, int target_Motor1_Command, int target_Motor2_Brake_command, int target_Motor2_Command, int target_FlipperMotor_Brake_command, int target_FlipperMotor_Command, int target_LACT_Command)
{
  Serial.println("--------- NEW TARGET Motor Values -----------");
  Serial.print("target_Motor1_Command: ");
  Serial.print(target_Motor1_Brake_command);
  Serial.print(":");
  Serial.println(target_Motor1_Command);
  Serial.print("target_Motor2_Command: ");
  Serial.print(target_Motor2_Brake_command);
  Serial.print(":");
  Serial.println(target_Motor2_Command);      
  Serial.print("target_FlipperMotor_Command: ");
  Serial.print(target_FlipperMotor_Brake_command);
  Serial.print(":");
  Serial.println(target_FlipperMotor_Command);      
  Serial.print("target_LACT_Command: ");
  Serial.println(target_LACT_Command);      
  Serial.println("---------------------------------------");

  //-----------------------------------
  //Set Brake Commands
  if (target_Motor1_Brake_command == 0)
    digitalWrite(TRED_L_MOTOR_BRAKE_PIN, HIGH);
  else
    digitalWrite(TRED_L_MOTOR_BRAKE_PIN, LOW);
        
  if (target_Motor2_Brake_command == 0)
    digitalWrite(TRED_R_MOTOR_BRAKE_PIN, HIGH);
  else
    digitalWrite(TRED_R_MOTOR_BRAKE_PIN, LOW);
        
  if (target_FlipperMotor_Brake_command == 0)
    digitalWrite(FLIPPER_MOTOR_BRAKE_PIN, HIGH);
  else
    digitalWrite(FLIPPER_MOTOR_BRAKE_PIN, LOW);
        
  //-----------------------------------
  //Set Motor Directions
  if (target_Motor1_Command < 0)
  {
    motor1backward(); //SWITCH ME TO BACKWARDS: refactor
  }
  else
    motor1forward(); //SWITCH ME TO FORWARDS: refactor

  if (target_Motor2_Command < 0)
  {
    motor2backward();//SWITCH ME TO BACKWARDS: refactor
  }
  else
    motor2forward();//SWITCH ME TO FORWARDS: refactor

  if (target_FlipperMotor_Command < 0)
    flipperMotor_backward();
  else
    flipperMotor_forward();  

    //Serial.print("FLIPPERMOTORCOMMAND: ********* ");
    //Serial.print(target_FlipperMotor_Command);
    //Serial.println("********* ");
    
  //-----------------------------------
  //Send the Speed Control Signal

  if ((target_Motor1_Command < 0 && target_Motor2_Command >= 0) ||
      (target_Motor1_Command < 0 && target_Motor2_Command >= 0))
  {
    //isRotating ... allow rotations to happen 
    analogWrite(TRED_L_MOTOR_SPEED_PIN, 255-abs(target_Motor1_Command));
    analogWrite(TRED_R_MOTOR_SPEED_PIN, 255-abs(target_Motor2_Command));

  }
  else
  {
    //Not in a rotation... block any of the treds from crashing into something
    if ((SAFTEY_SHUTDOWN_BACKWARD == 0 && target_Motor1_Command < 0) ||
      (SAFTEY_SHUTDOWN_FORWARD == 0 && target_Motor1_Command >= 0))
      analogWrite(TRED_L_MOTOR_SPEED_PIN, 255-abs(target_Motor1_Command));
  else
      analogWrite(TRED_L_MOTOR_SPEED_PIN, 255);

  if ((SAFTEY_SHUTDOWN_BACKWARD == 0 && target_Motor2_Command < 0) ||
      (SAFTEY_SHUTDOWN_FORWARD == 0 && target_Motor2_Command >= 0))
      analogWrite(TRED_R_MOTOR_SPEED_PIN, 255-abs(target_Motor2_Command));
  else
      analogWrite(TRED_R_MOTOR_SPEED_PIN, 255);

  }
  
  
  analogWrite(FLIPPER_MOTOR_SPEED_PIN, 255-abs(target_FlipperMotor_Command));
  setMotorSpeed(target_LACT_Command);
  //-----------------------------------
}

void releaseBrake()
{
  Serial.println( "***** BRAKE IS OFF!!! *****" );
  digitalWrite(TRED_L_MOTOR_BRAKE_PIN, HIGH);
  digitalWrite(TRED_R_MOTOR_BRAKE_PIN, HIGH);
}


void brake()
{
  Serial.println( "***** BRAKE IS ON!!! *****" );
  digitalWrite(TRED_L_MOTOR_BRAKE_PIN, LOW);
  digitalWrite(TRED_R_MOTOR_BRAKE_PIN, LOW);
}


void backward()
{
  Serial.println( "BACKWARD" );
  digitalWrite(TRED_L_MOTOR_DIRECTION_PIN, HIGH); // (towards back direction)
  digitalWrite(TRED_R_MOTOR_DIRECTION_PIN, LOW); // (towards back direction)
}


void motor1backward()
{
  Serial.println( "M1 FORWARD" );
  
  digitalWrite(TRED_L_MOTOR_DIRECTION_PIN, LOW); // (towards back direction)
}


void motor1forward()
{
  Serial.println( "M1 BACKWARD" );
  
  digitalWrite(TRED_L_MOTOR_DIRECTION_PIN, HIGH); // (towards front direction)
}


void motor2backward()
{
  Serial.println( "M2 FORWARD" );
  
  digitalWrite(TRED_R_MOTOR_DIRECTION_PIN, HIGH); // (towards back direction)
}


void motor2forward()
{
  Serial.println( "M2 BACKWARD" );
  
  digitalWrite(TRED_R_MOTOR_DIRECTION_PIN, LOW); // (towards front direction)
}

void flipperMotor_forward()
{
  Serial.println( "FLIPPER FORWARD" );
  
  digitalWrite(FLIPPER_MOTOR_DIRECTION_PIN, LOW); // (towards front direction)
}

void flipperMotor_backward()
{
  Serial.println( "FLIPPER BACKWARD" );
  
  digitalWrite(FLIPPER_MOTOR_DIRECTION_PIN, HIGH); // (towards back direction)
}

//------ IMU Functions -----

void initialize_IMU()
{
  // Read the WHO_AM_I register, this is a good test of communication
  byte c = myIMU.readByte(MPU9250_ADDRESS, WHO_AM_I_MPU9250);
  Serial.print(F("MPU9250 I AM 0x"));
  Serial.print(c, HEX);
  Serial.print(F(" I should be 0x"));
  Serial.println(0x71, HEX);

  if (c == 0x71) // WHO_AM_I should always be 0x71
  {
    Serial.println(F("MPU9250 is online..."));
    myIMU.calibrateMPU9250(myIMU.gyroBias, myIMU.accelBias);
    myIMU.initMPU9250();
    myIMU.initAK8963(myIMU.factoryMagCalibration);
    myIMU.getAres();
    myIMU.getGres();
    myIMU.getMres();
    Serial.println("MPU9250 initialized for active data mode....");
  } // if (c == 0x71)
  else
  {
    Serial.print("Could not connect to MPU9250: 0x");
    Serial.println(c, HEX);

    // Communication failed, stop here
    Serial.println(F("Communication failed, abort!"));
    Serial.flush();
    abort();
  }
}


void processIMU()
{
    // This is a hardware like bitwise speed choke to turn MHz into kilo hertz.... or nanoseconds into microseconds
    IMU_Processing_Index1++;
    if (IMU_Processing_Index1 > IMU_PROCESSING_SPEED1)
    {
       IMU_Processing_Index1 = 0;
       IMU_Processing_Index2++;
    }
    if (IMU_Processing_Index2 > IMU_PROCESSING_SPEED2)
    {
       IMU_Processing_Index2 = 0;
       Serial.println("IMU Pulse");
       calculateIMU();
    }
}


void calculateIMU()
{
  if (myIMU.readByte(MPU9250_ADDRESS, INT_STATUS) & 0x01)
  {
    myIMU.readAccelData(myIMU.accelCount);  // Read the x/y/z adc values
    myIMU.ax = (float)myIMU.accelCount[0] * myIMU.aRes; // - myIMU.accelBias[0];
    myIMU.ay = (float)myIMU.accelCount[1] * myIMU.aRes; // - myIMU.accelBias[1];
    myIMU.az = (float)myIMU.accelCount[2] * myIMU.aRes; // - myIMU.accelBias[2];

    myIMU.readGyroData(myIMU.gyroCount);  // Read the x/y/z adc values
    myIMU.gx = (float)myIMU.gyroCount[0] * myIMU.gRes;
    myIMU.gy = (float)myIMU.gyroCount[1] * myIMU.gRes;
    myIMU.gz = (float)myIMU.gyroCount[2] * myIMU.gRes;

    myIMU.readMagData(myIMU.magCount);  // Read the x/y/z adc values
    myIMU.mx = (float)myIMU.magCount[0] * myIMU.mRes
               * myIMU.factoryMagCalibration[0] - myIMU.magBias[0];
    myIMU.my = (float)myIMU.magCount[1] * myIMU.mRes
               * myIMU.factoryMagCalibration[1] - myIMU.magBias[1];
    myIMU.mz = (float)myIMU.magCount[2] * myIMU.mRes
               * myIMU.factoryMagCalibration[2] - myIMU.magBias[2];
  }

  myIMU.updateTime();

  MahonyQuaternionUpdate(myIMU.ax, myIMU.ay, myIMU.az, myIMU.gx * DEG_TO_RAD,
                         myIMU.gy * DEG_TO_RAD, myIMU.gz * DEG_TO_RAD, myIMU.my,
                         myIMU.mx, myIMU.mz, myIMU.deltat);

  myIMU.delt_t = millis() - myIMU.count;
  if (myIMU.delt_t > 100)
  {
    ///*
    Serial.print("ax = ");  Serial.print((int)1000 * myIMU.ax);
    Serial.print(" ay = "); Serial.print((int)1000 * myIMU.ay);
    Serial.print(" az = "); Serial.print((int)1000 * myIMU.az);
    Serial.println(" mg");

    Serial.print("gx = ");  Serial.print(myIMU.gx, 2);
    Serial.print(" gy = "); Serial.print(myIMU.gy, 2);
    Serial.print(" gz = "); Serial.print(myIMU.gz, 2);
    Serial.println(" deg/s");

    Serial.print("mx = ");  Serial.print((int)myIMU.mx);
    Serial.print(" my = "); Serial.print((int)myIMU.my);
    Serial.print(" mz = "); Serial.print((int)myIMU.mz);
    Serial.println(" mG");

    Serial.print("q0 = ");  Serial.print(*getQ());
    Serial.print(" qx = "); Serial.print(*(getQ() + 1));
    Serial.print(" qy = "); Serial.print(*(getQ() + 2));
    Serial.print(" qz = "); Serial.println(*(getQ() + 3));
    //*/
    myIMU.yaw   = atan2(2.0f * (*(getQ()+1) * *(getQ()+2) + *getQ()
                  * *(getQ()+3)), *getQ() * *getQ() + *(getQ()+1)
                  * *(getQ()+1) - *(getQ()+2) * *(getQ()+2) - *(getQ()+3)
                  * *(getQ()+3));
    myIMU.pitch = -asin(2.0f * (*(getQ()+1) * *(getQ()+3) - *getQ()
                  * *(getQ()+2)));
    myIMU.roll  = atan2(2.0f * (*getQ() * *(getQ()+1) + *(getQ()+2)
                  * *(getQ()+3)), *getQ() * *getQ() - *(getQ()+1)
                  * *(getQ()+1) - *(getQ()+2) * *(getQ()+2) + *(getQ()+3)
                  * *(getQ()+3));
    myIMU.pitch *= RAD_TO_DEG;
    myIMU.yaw   *= RAD_TO_DEG;
    myIMU.yaw  -= 3.75; //True North for Austin, TX
    myIMU.roll *= RAD_TO_DEG;
    ///*
    Serial.print("Yaw, Pitch, Roll: ");
    Serial.print(myIMU.yaw, 2);
    Serial.print(", ");
    Serial.print(myIMU.pitch, 2);
    Serial.print(", ");
    Serial.println(myIMU.roll, 2);
    //*/
    myIMU.tempCount = myIMU.readTempData();  // Read the adc values
    // Temperature in degrees Centigrade
    myIMU.temperature = ((float) myIMU.tempCount) / 333.87 + 21.0;
    // Print temperature in degrees Centigrade
    ///*
    Serial.print("Temperature is ");  Serial.print(myIMU.temperature, 1);
    Serial.println(" degrees C");
    
    Serial.print("rate = ");
    Serial.print((float)myIMU.sumCount / myIMU.sum, 2);
    Serial.println(" Hz");
    //*/
    myIMU.count = millis();
    myIMU.sumCount = 0;
    myIMU.sum = 0;
  }
}


//------ LACT Functions -----

void exitSafeStart()
{
  smcSerial.write(0x83);
}
 
// speed should be a number from -3200 to 3200
void setMotorSpeed(int speed)
{
  //Piggyback and reset the LACT if we are having problems
  if (speed == -3201)
  {
    exitSafeStart();
    speed = -3200;
  }
  if (speed == 3201)
  {
    exitSafeStart();
    speed = 3200;
  }
  
  if (speed < 0)
  {
    smcSerial.write(0x86);  // motor reverse command
    speed = -speed;  // make speed positive
  }
  else
  {
    smcSerial.write(0x85);  // motor forward command
  }
  smcSerial.write(speed & 0x1F);
  smcSerial.write(speed >> 5);
}
 
